@@ -57,7 +57,7 @@ functions, where the component functions have these signatures:
 
   Returns:
     A pytree representing the parameters extracted from `opt_state`, such that
-    the invariant `params == get_params(init_params(params))` holds true.
+    the invariant `params == get_params(init_fun(params))` holds true.
 
 
 Notice that an optimizer implementation has a lot of flexibility in the form of
@@ -66,16 +66,11 @@ the JAX transforms defined in api.py) and it has to be consumable by update_fun
 and get_params.
 """
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
 
 from collections import namedtuple
 import functools
 import operator
 
-from six.moves import reduce
-
 import jax.numpy as np
 from jax.util import partial, safe_zip, safe_map, unzip2
 from jax import tree_util
@@ -201,11 +196,12 @@ def sgd(step_size):
 
 @optimizer
 def momentum(step_size, mass):
-  """Construct optimizer triple for SGD with Nesterov momentum.
+  """Construct optimizer triple for SGD with momentum.
 
   Args:
     step_size: positive scalar, or a callable representing a step size schedule
       that maps the iteration index to positive scalar.
+    mass: positive scalar representing the momentum coefficient.
 
   Returns:
     An (init_fun, update_fun, get_params) triple.
@@ -216,8 +212,35 @@ def momentum(step_size, mass):
     return x0, v0
   def update(i, g, state):
     x, velocity = state
-    velocity = mass * velocity - (1. - mass) * g
-    x = x + step_size(i) * velocity
+    velocity = mass * velocity + g
+    x = x - step_size(i) * velocity
+    return x, velocity
+  def get_params(state):
+    x, _ = state
+    return x
+  return init, update, get_params
+
+
+@optimizer
+def nesterov(step_size, mass):
+  """Construct optimizer triple for SGD with Nesterov momentum.
+
+  Args:
+    step_size: positive scalar, or a callable representing a step size schedule
+      that maps the iteration index to positive scalar.
+    mass: positive scalar representing the momentum coefficient.
+
+  Returns:
+    An (init_fun, update_fun, get_params) triple.
+  """
+  step_size = make_schedule(step_size)
+  def init(x0):
+    v0 = np.zeros_like(x0)
+    return x0, v0
+  def update(i, g, state):
+    x, velocity = state
+    velocity = mass * velocity + g
+    x = x - step_size(i) * (mass * velocity + g)
     return x, velocity
   def get_params(state):
     x, _ = state
@@ -269,24 +292,61 @@ def rmsprop(step_size, gamma=0.9, eps=1e-8):
   Args:
     step_size: positive scalar, or a callable representing a step size schedule
       that maps the iteration index to positive scalar.
+      gamma: Decay parameter.
+      eps: Epsilon parameter.
 
   Returns:
     An (init_fun, update_fun, get_params) triple.
   """
   step_size = make_schedule(step_size)
   def init(x0):
-    avg_sq_grad = np.ones_like(x0)
+    avg_sq_grad = np.zeros_like(x0)
     return x0, avg_sq_grad
   def update(i, g, state):
     x, avg_sq_grad = state
     avg_sq_grad = avg_sq_grad * gamma + g**2 * (1. - gamma)
-    x = x - step_size(i) * g / (np.sqrt(avg_sq_grad) + eps)
+    x = x - step_size(i) * g / np.sqrt(avg_sq_grad + eps)
     return x, avg_sq_grad
   def get_params(state):
     x, _ = state
     return x
   return init, update, get_params
 
+
+@optimizer
+def rmsprop_momentum(step_size, gamma=0.9, eps=1e-8, momentum=0.9):
+  """Construct optimizer triple for RMSProp with momentum.
+
+  This optimizer is separate from the rmsprop optimizer because it needs to
+  keep track of additional parameters.
+
+  Args:
+    step_size: positive scalar, or a callable representing a step size schedule
+      that maps the iteration index to positive scalar.
+    gamma: Decay parameter.
+    eps: Epsilon parameter.
+    momentum: Momentum parameter.
+
+  Returns:
+    An (init_fun, update_fun, get_params) triple.
+  """
+  step_size = make_schedule(step_size)
+  def init(x0):
+    avg_sq_grad = np.zeros_like(x0)
+    mom = np.zeros_like(x0)
+    return x0, avg_sq_grad, mom
+  def update(i, g, state):
+    x, avg_sq_grad, mom = state
+    avg_sq_grad = avg_sq_grad * gamma + g**2 * (1. - gamma)
+    mom = momentum * mom + step_size(i) * g / np.sqrt(avg_sq_grad + eps)
+    x = x - mom
+    return x, avg_sq_grad, mom
+  def get_params(state):
+    x, _, _ = state
+    return x
+  return init, update, get_params
+
+
 @optimizer
 def adam(step_size, b1=0.9, b2=0.999, eps=1e-8):
   """Construct optimizer triple for Adam.
@@ -355,7 +415,7 @@ def sm3(step_size, momentum=0.9):
   def update(i, g, state):
     x, m, vs = state
     vs = [broadcast_into(g.ndim, v, i) for i, v in enumerate(vs)]
-    accum = reduce(np.minimum, vs) + g ** 2
+    accum = functools.reduce(np.minimum, vs) + g ** 2
     accum_inv_sqrt = np.where(accum > 0, 1. / np.sqrt(accum), 0)
     m = (1. - momentum) * (g * accum_inv_sqrt) + momentum * m
     x = x - step_size(i) * m
@@ -390,6 +450,14 @@ def inverse_time_decay(step_size, decay_steps, decay_rate, staircase=False):
       return step_size / (1 + decay_rate * i / decay_steps)
   return schedule
 
+def polynomial_decay(step_size, decay_steps, final_step_size, power=1.0):
+  def schedule(step_num):
+    step_num = np.minimum(step_num, decay_steps)
+    step_mult = (1 - step_num / decay_steps) ** power
+    return step_mult * (step_size - final_step_size) + final_step_size
+
+  return schedule
+
 def piecewise_constant(boundaries, values):
   boundaries = np.array(boundaries)
   values = np.array(values)
@@ -423,3 +491,52 @@ def clip_grads(grad_tree, max_norm):
   norm = l2_norm(grad_tree)
   normalize = lambda g: np.where(norm < max_norm, g, g * (max_norm / norm))
   return tree_map(normalize, grad_tree)
+
+
+### serialization utilities
+
+class JoinPoint(object):
+  """Marks the boundary between two joined (nested) pytrees."""
+  def __init__(self, subtree):
+    self.subtree = subtree
+
+  # Since pytrees are containers of numpy arrays, look iterable.
+  def __iter__(self):
+    yield self.subtree
+
+def unpack_optimizer_state(opt_state):
+  """Converts an OptimizerState to a marked pytree.
+
+  Converts an OptimizerState to a marked pytree with the leaves of the outer
+  pytree represented as JoinPoints to avoid losing information. This function is
+  intended to be useful when serializing optimizer states.
+
+  Args:
+    opt_state: An OptimizerState
+  Returns:
+    A pytree with JoinPoint leaves that contain a second level of pytrees.
+  """
+  packed_state, tree_def, subtree_defs = opt_state
+  subtrees = map(tree_unflatten, subtree_defs, packed_state)
+  sentinels = [JoinPoint(subtree) for subtree in subtrees]
+  return tree_util.tree_unflatten(tree_def, sentinels)
+
+def pack_optimizer_state(marked_pytree):
+  """Converts a marked pytree to an OptimizerState.
+
+  The inverse of unpack_optimizer_state. Converts a marked pytree with the
+  leaves of the outer pytree represented as JoinPoints back into an
+  OptimizerState. This function is intended to be useful when deserializing
+  optimizer states.
+
+  Args:
+    marked_pytree: A pytree containing JoinPoint leaves that hold more pytrees.
+  Returns:
+    An equivalent OptimizerState to the input argument.
+  """
+  sentinels, tree_def = tree_flatten(marked_pytree)
+  assert all(isinstance(s, JoinPoint) for s in sentinels)
+  subtrees = [s.subtree for s in sentinels]
+  states_flat, subtree_defs = unzip2(map(tree_flatten, subtrees))
+  packed_state = pack(map(pack, states_flat))
+  return OptimizerState(packed_state, tree_def, subtree_defs)
