@@ -14,7 +14,6 @@
 # limitations under the License.
 
 
-from functools import partial
 import numpy as onp
 
 from jax.numpy import lax_numpy as np
@@ -71,9 +70,15 @@ def svd(x, full_matrices=True, compute_uv=True):
 def triangular_solve(a, b, left_side=False, lower=False, transpose_a=False,
                      conjugate_a=False, unit_diagonal=False):
   conjugate_a = conjugate_a and np.issubdtype(lax.dtype(a), np.complexfloating)
-  return triangular_solve_p.bind(
+  singleton = np.ndim(b) == np.ndim(a) - 1
+  if singleton:
+    b = np.expand_dims(b, -1 if left_side else -2)
+  out = triangular_solve_p.bind(
       a, b, left_side=left_side, lower=lower, transpose_a=transpose_a,
       conjugate_a=conjugate_a, unit_diagonal=unit_diagonal)
+  if singleton:
+    out = out[..., 0] if left_side else out[..., 0, :]
+  return out
 
 
 # utilities
@@ -491,7 +496,7 @@ def _lu_unblocked(a):
   return lax.fori_loop(0, min(m, n), body, (pivot, perm, a))
 
 
-def _lu_blocked(a, block_size=32):
+def _lu_blocked(a, block_size=128):
   """Blocked LU decomposition, as an unrolled loop."""
   m, n = a.shape
   r = min(m, n)
@@ -499,13 +504,12 @@ def _lu_blocked(a, block_size=32):
   for k in range(0, r, block_size):
     b = min(r - k, block_size)
     block_pivot, perm, lu_block = _lu_unblocked(a[k:, k:k+b])
-    a = ops.index_update(a, ops.index[k:, k:k+b], lu_block)
 
-    a = ops.index_update(a, ops.index[k:, :k], a[perm + k, :k])
+    a = ops.index_update(a, ops.index[k:, :], a[perm + k, :])
+    a = ops.index_update(a, ops.index[k:, k:k+b], lu_block)
     pivot = ops.index_update(pivot, ops.index[k:k+b], block_pivot + k)
 
     if k + b < n:
-      a = ops.index_update(a, ops.index[k:, k+b:], a[perm + k, k+b:])
       a = ops.index_update(
         a, ops.index[k:k+b, k+b:],
         triangular_solve(a[k:k+b, k:k+b], a[k:k+b, k+b:],
