@@ -56,6 +56,9 @@ acos = onp.arccos
 atan = onp.arctan
 sinh = onp.sinh
 cosh = onp.cosh
+asinh = onp.arcsinh
+acosh = onp.arccosh
+atanh = onp.arctanh
 
 betainc = scipy.special.betainc
 lgamma = scipy.special.gammaln
@@ -263,7 +266,7 @@ sort = onp.sort
 def sort_key_val(keys, values, dimension=-1):
   idxs = list(onp.ix_(*[onp.arange(d) for d in keys.shape]))
   idxs[dimension] = onp.argsort(keys, axis=dimension)
-  return keys[idxs], values[idxs]
+  return keys[tuple(idxs)], values[tuple(idxs)]
 
 # TODO untake
 
